/* Flow.js related elements */
/*noinspection CssUnusedSymbol*/
.flow-drop {
    padding:15px;
    font-size:13px;
    text-align:center;
    color:#666;
    font-weight:bold;
    background-color:#eee;
    border: 2px dashed #aaa;
    border-radius:10px;
    margin-top:40px;
    z-index:9999;
}

/*noinspection CssUnusedSymbol*/
.flow-dragover {
    padding:30px;
    color:#555;
    background-color:#ddd;
    border:1px solid #999;
}

/*noinspection CssUnusedSymbol*/
.flow-browse {
    text-decoration: underline;
}

/*noinspection CssUnusedSymbol*/
.flow-error {
    font-size:14px;
    font-style:italic;
}

/*noinspection CssUnusedSymbol*/
.flow-progress {
    /*margin:30px 0 30px 0;*/
    width:100%;
}
